@@ -1117,11 +1117,11 @@ next:
 		path.slots[0]++;
 	}
 
-	if (restore_metadata) {
+	if ((restore_metadata || get_xattrs) && !dry_run) {
 		snprintf(path_name, PATH_MAX, "%s%s", output_rootdir, in_dir);
 		fd = open(path_name, O_RDONLY);
 		if (fd < 0) {
-			error("failed to access '%s' to restore metadata: %m",
+			error("failed to access '%s' to restore metadata/xattrs: %m",
 					path_name);
 			if (!ignore_errors) {
 				ret = -1;
@@ -1132,7 +1132,21 @@ next:
 			 * Set owner/mode/time on the directory as well
 			 */
 			key->type = BTRFS_INODE_ITEM_KEY;
-			ret = copy_metadata(root, fd, key);
+			if (restore_metadata) {
+				ret = copy_metadata(root, fd, key);
+				if (ret && !ignore_errors) {
+					close(fd);
+					goto out;
+				}
+			}
+
+			/* Also set xattrs on the directory. */
+			if (get_xattrs) {
+				ret = set_file_xattrs(root, key->objectid, fd, path_name);
+				if (ret) {
+					error("failed to set xattrs on %s: %m", path_name);
+				}
+			}
 			close(fd);
 			if (ret && !ignore_errors)
 				goto out;
@@ -1333,31 +1347,35 @@ static const char * const cmd_restore_usage[] = {
 	"btrfs restore [options] -l <device>",
 	"Try to restore files from a damaged filesystem (unmounted)",
 	"",
-	"  control:",
-	"       -D|--dry-run         dry run (only list files that would be recovered)",
-	"       -i|--ignore-errors   ignore errors",
-	"       -o|--overwrite       overwrite",
-	"  restoration:",
-	"       -m|--metadata        restore owner, mode and times",
-	"       -S|--symlink         restore symbolic links",
-	"       -s|--snapshots       get snapshots",
-	"       -x|--xattr           restore extended attributes",
-	"  filtering:",
-	"       --path-regex <regex>",
-	"                            restore only filenames matching regex,",
-	"                            you have to use following syntax (possibly quoted):",
-	"                            ^/(|home(|/username(|/Desktop(|/.*))))$",
-	"       -c                   ignore case (--path-regex only)",
-	"  analysis:",
-	"       -d                   find dir",
-	"       -l|--list-roots      list tree roots",
-	"  alternate starting point:",
-	"       -f <bytenr>          filesystem location",
-	"       -r|--root <rootid>   root objectid",
-	"       -t <bytenr>          tree location",
-	"       -u|--super <mirror>  super mirror",
-	"  other:",
-	"       -v|--verbose         deprecated, alias for global -v option",
+	"Control:",
+	OPTLINE("-D|--dry-run", "dry run (only list files that would be recovered)"),
+	OPTLINE("-i|--ignore-errors", "ignore errors"),
+	OPTLINE("-o|--overwrite", "overwrite"),
+	"",
+	"Restoration:",
+	OPTLINE("-m|--metadata", "restore owner, mode and times"),
+	OPTLINE("-S|--symlink", "restore symbolic links"),
+	OPTLINE("-s|--snapshots", "get snapshots"),
+	OPTLINE("-x|--xattr", "restore extended attributes"),
+	"",
+	"Filtering:",
+	OPTLINE("--path-regex <regex>", "restore only filenames matching regex, "
+		"you have to use following syntax (possibly quoted): "
+		"^/(|home(|/username(|/Desktop(|/.*))))$"),
+	OPTLINE("-c", "ignore case (--path-regex only)"),
+	"",
+	"Analysis:",
+	OPTLINE("-d", "find dir"),
+	OPTLINE("-l|--list-roots", "list tree roots"),
+	"",
+	"Alternate starting point:",
+	OPTLINE("-f <bytenr>", "filesystem location"),
+	OPTLINE("-r|--root <rootid>", "root objectid"),
+	OPTLINE("-t <bytenr>", "tree location"),
+	OPTLINE("-u|--super <mirror>", "super mirror"),
+	"",
+	"Other:",
+	OPTLINE("-v|--verbose", "deprecated, alias for global -v option"),
 	HELPINFO_INSERT_GLOBALS,
 	HELPINFO_INSERT_VERBOSE,
 	"",
@@ -1481,9 +1499,9 @@ static int cmd_restore(const struct cmd_struct *cmd, int argc, char **argv)
 	}
 
 	if (!list_roots && check_argc_min(argc - optind, 2))
-		usage(cmd);
+		usage(cmd, 1);
 	else if (list_roots && check_argc_min(argc - optind, 1))
-		usage(cmd);
+		usage(cmd, 1);
 
 	if (fs_location && root_objectid) {
 		error("can't use -f and -r at the same time");
